/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: i2c_buffer.c,v 1.6 2005-04-27 19:23:00 zer0 Exp $
 *
 */

#include <compat/twi.h>

#include <list/list.h>

#include <i2c/i2cs_genric.h>
#include <i2c/i2cm_genric.h>

#define DEBUG 0
#include <stdio.h>

#define debug_printf(args...) do { if(DEBUG) printf(args); } while(0)


#include <i2c_config.h>

/* if we used bufferized i2c define this code */
#ifdef I2C_BUFFERIZED

#if (!(defined I2C_MASTER) && !(defined I2C_SLAVE)) || ((defined I2C_MASTER) && (defined I2C_SLAVE))
#error You must define I2C_MASTER OR I2C_SLAVE in i2c_config.h
#endif

extern volatile u08 received_data;
typedef enum { I2C_BUFFER_READY, I2C_BUFFER_BUSY, I2C_BUFFER_SCAN} state_i2c_buffer; 

/* Type defs */
LIST_TYPEDEF(i2c_recv_fifo_t, u08, I2C_RECV_FIFO_SIZE);
LIST_TYPEDEF(i2c_send_fifo_t, u08, I2C_SEND_FIFO_SIZE);

/** i2c state */
static state_i2c_buffer g_state = I2C_BUFFER_READY;

/* fifos */
static i2c_recv_fifo_t g_recv;
static i2c_send_fifo_t g_send;

/** recv event, called when we receive a frame
   params are : data buffer and size */
static void (*g_recv_event)(u08 *, u08) = NULL;

/** send event, called when transmit is complete 
 param is error code : 0 if success */
static void (*g_send_event)(u08) = NULL;


/* ************** MASTER ************ */
#ifdef I2C_MASTER
void master_event(i2cm_state s)
{
  char c;
  static u08 size=0;

  switch(s) {
  case I2CM_SENT_START :
    debug_printf("st\r\n");
    if(g_state == I2C_BUFFER_BUSY) {
      LIST_READ(g_send, &c);
      /* FIXME : get return value of sendbyte*/ 
      i2cm_send_byte(c);
    }
    else { /* I2C_BUFFER_SCAN */
      i2cm_get_received_byte();
      g_state = I2C_BUFFER_BUSY;
      size=0;
    }
    break;
      
  case I2CM_RECEIVED_BYTE:
    if(!size) { /* this is the first byte ==> size */
      if(received_data == 0) {
	i2cm_send_stop();
	g_state = I2C_BUFFER_READY;
      }
      else {
	size=i2cm_get_received_byte();
      }
    }
    else if(size!=1) { /* intermediate data */
      c=i2cm_get_received_byte();
      LIST_PUSH_END(g_recv, c);
      debug_printf("1%c",c);
      size--;
    }
    else {
      void (*event)(u08 *, u08);
      
      c=i2cm_send_stop();
      LIST_PUSH_END(g_recv, c);
      debug_printf("2%c",c);
      
      IRQ_LOCK();
      event = g_recv_event;
      IRQ_UNLOCK();
      /* FIXME : implement a LIST2PTR function */
      if(event) event(&g_recv.elt[g_recv.hdr.beg_indice], LIST_CURSIZE(g_recv));
      g_state = I2C_BUFFER_READY;
    }
    break;
    
    
  case I2CM_SENT_BYTE :
    /* move list cursor right, test is true if end of list */
    if(LIST_READ_MOVE(g_send, &c, 1)) {
      void (*event)(u08);
      IRQ_LOCK();
      event = g_send_event;
      IRQ_UNLOCK();
      i2cm_send_stop();
      g_state = I2C_BUFFER_READY;
      if(event) event(0);
    }
    else {
      /* FIXME : get return value of sendbyte*/ 
      i2cm_send_byte(c);
    }
    break;

  default: 
    { /* handle error (call send_event with error code =1) */
/*        void (*event)(u08); */
/*        IRQ_LOCK(); */
/*        event = g_send_event; */
/*        IRQ_UNLOCK(); */
    debug_printf("err\r\n");
    }
    break;
  }
}

/** send a buffer, return 0 if transmission starts */
u08 i2c_send(u08 slave, u08 * data, u08 size)
{
  /* FIXME : define error codes */
  /* i2c ready ? */
  IRQ_LOCK();
  if ( g_state != I2C_BUFFER_READY ) {
    IRQ_UNLOCK();
    return 1;
  }

  if(size==0) {
    IRQ_UNLOCK();
    return 2;
  }

  /* fill fifo */
  LIST_INIT(g_send, 0);
  if(LIST_ARRAY_PUSH_START(g_send, data, size)) {
    IRQ_UNLOCK();
    return 3;
  }

  /* update state */
  g_state = I2C_BUFFER_BUSY;
  
  LIST_READ_START(g_send,(char *) NULL);
  
  /* start xmit */
  i2cm_send_start(slave<<1 | TW_WRITE);
  
  IRQ_UNLOCK();
  return 0;
}

/* query slave */
u08 i2c_scan(u08 slave)
{
  /* FIXME : define error codes */
  /* i2c ready ? */
  IRQ_LOCK();
  if ( g_state != I2C_BUFFER_READY ) {
    IRQ_UNLOCK();
    return 1;
  }
  
  LIST_INIT(g_recv, 0);
  g_state = I2C_BUFFER_SCAN;
  i2cm_send_start(slave<<1 | TW_READ);
  IRQ_UNLOCK();
  return 0;
}

void i2c_buffer_init(void)
{
  LIST_INIT(g_recv, 0);
  LIST_INIT(g_send, 0);
  i2cm_register_event( master_event ) ;
}
#endif


/* ************** SLAVE ************ */
#ifdef I2C_SLAVE
void slave_event(i2cs_state s)
{
  u08 c;

  switch(s) {
  case I2CS_ADRESSED_SEND:
    g_state = I2C_BUFFER_BUSY;
    /* there is something in the fifo */
    if(LIST_CURSIZE(g_send)) {
      i2cs_send_byte(LIST_CURSIZE(g_send));
    }
    else {
      /* else we send 0, which means nothing to emit */
      i2cs_send_last_byte(0);
    }
    break;
    
  case I2CS_SENT_BYTE_NO_ACK:
    debug_printf("3");
    //    debug_printf("noack\r\n");
    // FIXME : handle error 
  case I2CS_SENT_BYTE: 
    LIST_READ(g_send, &c);
    if(LIST_READ_MOVE(g_send, (char *)NULL, 1)) {
      void (*event)(u08);
      debug_printf("4");
      
      i2cs_send_last_byte(c);
      IRQ_LOCK();
      event = g_send_event;
      IRQ_UNLOCK();
      if(event) event(0);
      LIST_INIT(g_send, 0);
    }
    else {
      debug_printf("5");
      i2cs_send_byte(c);
    }
    break;
    
    
  case I2CS_RECEIVED_BYTE:
    debug_printf("6");
    /* add char to the list */
    LIST_PUSH_END(g_recv, i2cs_get_received_byte());
    break;

    /* recv finished */
  case I2CS_READY:
    {
      debug_printf("7");
      void (*event)(u08 *, u08);
      IRQ_LOCK();
      event = g_recv_event;
      IRQ_UNLOCK();
      /* FIXME : implement a LIST2PTR function */
      if(event) event(&g_recv.elt[g_recv.hdr.beg_indice], LIST_CURSIZE(g_recv));
    }
    break;

  case I2CS_ADRESSED_RECEIVE:
      debug_printf("8");
    /* empty the list, and initialize start/end cursors at the beginning */
    LIST_INIT(g_recv, 0);
    break;
    
  default:
    break;
  }
}


/** send a buffer, return 0 if success. the function will wait a
    master scan */
u08 i2c_send(u08 * data, u08 size)
{
  /* FIXME : define error codes */
  /* i2c ready ? */
  IRQ_LOCK();
  if ( LIST_CURSIZE(g_send) ) {
    IRQ_UNLOCK();
    return 1;
  }

  if(size==0) {
    IRQ_UNLOCK();
    return 2;
  }

  /* fill fifo */
  if(LIST_ARRAY_PUSH_START(g_send, data, size)) {
    IRQ_UNLOCK();
    return 3;
  }

  LIST_READ_START(g_send,(char *) NULL);
  
  IRQ_UNLOCK();
  return 0;
}


void i2c_buffer_init(void)
{
  LIST_INIT(g_recv, 0);
  LIST_INIT(g_send, 0);
  i2cs_register_event( slave_event ) ;
}
#endif



/* *************** COMMON ***************** */



/* register event when receiving a buffer */
void i2c_register_recv_event(void (*event)(u08 *, u08))
{
  IRQ_LOCK();
  g_recv_event = event ;
  IRQ_UNLOCK();
}

/* register event when transmit complete */
void i2c_register_send_event(void (*event)(u08))
{
}

#endif
