/*  
 *  Copyright Droids Corporation, Microb Technology, Eirbot (2005)
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: i2c0.c,v 1.11 2005-04-27 19:23:00 zer0 Exp $
 *
 */

#include <avr/signal.h>
#include <compat/twi.h>
#include <i2cs_genric.h>
#include <i2cm_genric.h>

#include <i2c_config.h>

#define NULL 0


volatile i2cm_state state_master = I2CM_NOT_INIT;
volatile i2cs_state state_slave  = I2CS_NOT_INIT;

volatile u08 received_data;
void (*event_master)(i2cm_state state) = NULL;
void (*event_slave)(i2cs_state state) = NULL;

/** master functions ********************************************************/

void i2cm_init(void)
{
  u08 flags;

  IRQ_LOCK(flags);
  
  TWCR = 0; // grace a ca, init peut servir de re-init
  TWBR = I2C_BITRATE;
  TWCR = (1<<TWEN) | (1<<TWIE) ;
  
  // prescaler
  if(I2C_PRESCALER & 1)
    sbi(TWSR, TWPS0);
  if(I2C_PRESCALER & 2)
    sbi(TWSR, TWPS1);
    
  state_master = I2CM_READY;

  IRQ_UNLOCK(flags);
}

void i2cm_manage(void)
{

}

void i2cm_register_event( void (*f)(i2cm_state state) ) 
{
  u08 flags;

  IRQ_LOCK(flags);
  event_master = f;
  IRQ_UNLOCK(flags);
}

i2cm_state i2cm_get_state(void)
{
  return state_master;
}

u08 i2cm_send_start(void)
{
  if ( (state_master == I2CM_READY) || (state_master == I2CM_NEW_BYTE_RECEIVED) || (state_master == I2CM_SENT_START_NO_ACK) || (state_master == I2CM_SENT_BYTE_NO_ACK) || (state_master == I2CM_SENT_BYTE) )
  {
    state_master = I2CM_BUSY;
    TWCR = (1<<TWSTA) | (1<<TWINT) | (1<<TWEN) | (1<<TWIE) ;
    return 0;
  }
  return 1;
}

u08 i2cm_send_stop(void)
{
  if ( (state_master == I2CM_NEW_BYTE_RECEIVED) || (state_master == I2CM_SENT_START_NO_ACK) || (state_master == I2CM_SENT_BYTE_NO_ACK) || (state_master == I2CM_SENT_BYTE) )
  {
    TWCR = (1<<TWSTO) | (1<<TWINT) | (1<<TWEN) | (1<<TWIE) ;
    state_master = I2CM_READY;
    return 0;
  }
  return 1;
}

u08 i2cm_get_received_byte(u08 *byte)
{
  if (state_master == I2CM_NEW_BYTE_RECEIVED)
  {
    *byte = received_data;
    return 0;
  }
  return 1;
}

u08 i2cm_want_last_byte(void)
{
  if ( (state_master == I2CM_SENT_START) || (state_master == I2CM_NEW_BYTE_RECEIVED) )
  {
    TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE); // don't ack !
    return 0;
  }
  return 1;
}

u08 i2cm_want_more_byte(void)
{
  if ( (state_master == I2CM_SENT_START) || (state_master == I2CM_NEW_BYTE_RECEIVED) )
  {
    TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE) | (1<<TWEA); // ack !
    return 0;
  }
  return 1;
}

u08 i2cm_send_byte( u08 c)
{
  if ( (state_master == I2CM_SENT_START) || (state_master == I2CM_SENT_START_NO_ACK) || (state_master == I2CM_SENT_BYTE) )
    {
      state_master = I2CM_BUSY;
      TWDR = c;
      TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE) ;
      return 0;
    }
  return 1;
}

u08 i2cm_send_address(u08 address)
{
  if (state_master == I2CM_INITIATED_START)
  {
    TWDR = address;
    TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE) ;
    return 0;
  }
  return 1;
}

void i2c_wait_for_ready(void)
{
  while(!(TWCR & (1<<TWINT)));
}


/** slave functions ********************************************************/

void i2cs_init( u08 slave_address)
{
  u08 flags;

  IRQ_LOCK(flags);
  
  TWBR = I2C_BITRATE;
  /* enable, enable int,  respond to own adress */
  TWCR = (1<<TWEA) | (1<<TWEN) | (1<<TWIE) ;
  
  // prescaler
  if(I2C_PRESCALER & 1)
    sbi(TWSR, TWPS0);
  if(I2C_PRESCALER & 2)
    sbi(TWSR, TWPS1);

  TWAR = slave_address << 1 ; // change for TWAR format
  
  // general call
  if (slave_address & 0x80)
    sbi(TWAR, TWGCE);


  state_slave = I2CS_READY; // flag

  IRQ_UNLOCK(flags);
}

void i2cs_manage(void)
{

}

void i2cs_register_event( void (*f)(i2cs_state state)  )
{
  u08 flags;
  IRQ_LOCK(flags);
  event_slave = f;
  IRQ_UNLOCK(flags);
}

i2cs_state i2cs_get_state(void)
{
  return state_slave;
}

u08 i2cs_get_received_byte(void)
{
  TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA) | (1<<TWIE) ;
  return received_data;
}

u08 i2cs_send_byte(u08 c)
{
/*    if ( (state_slave == I2CS_ADRESSED_SEND) || */
/*          (state_slave == I2CS_SENT_BYTE) || */
/*          (state_slave == I2CS_SENT_BYTE_NO_ACK) */
/*           )  */
    {
      state_slave = I2CS_BUSY;
      TWDR = c;
      TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE) | (1<<TWEA) ;
      return 0;
    }
  return 1;
}

u08 i2cs_send_last_byte(u08 c)
{
/*    if ( (state_slave == I2CS_ADRESSED_SEND) || */
/*          (state_slave == I2CS_SENT_BYTE) || */
/*          (state_slave == I2CS_SENT_BYTE_NO_ACK) */
/*           )  */
    {
      state_slave = I2CS_BUSY;
      TWDR = c;
      TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE) ;
      return 0;
    }
  return 1;
}


/** interrupt ********************************************************/

//INTERRUPT(SIG_2WIRE_SERIAL) // other ints enabled
SIGNAL(SIG_2WIRE_SERIAL) // other ints disabled
{
  switch(TW_STATUS)
    {
    /* master */

    case TW_START:      // a start has been transmitted, transmit SLA+W
    case TW_REP_START:
      state_master = I2CM_INITIATED_START;
      if (event_master != NULL) event_master(state_master);
      break;

    // SLA+W transmitted
    case TW_MT_SLA_ACK:
    case TW_MR_SLA_ACK:
      state_master = I2CM_SENT_START;
      if (event_master != NULL) event_master(state_master);
      break;
      
    case TW_MT_SLA_NACK:
    case TW_MR_SLA_NACK:
      state_master = I2CM_SENT_START_NO_ACK; 
      if (event_master != NULL) event_master(state_master);
      break;

    case TW_MT_DATA_ACK:
      state_master = I2CM_SENT_BYTE;
      if (event_master != NULL) event_master(state_master);
      break;

    case TW_MR_DATA_NACK:
    case TW_MR_DATA_ACK:
      received_data = TWDR;
      state_master = I2CM_NEW_BYTE_RECEIVED;
      if (event_master != NULL) event_master(state_master);
      break;

      
    case TW_MT_DATA_NACK:
      state_master = I2CM_SENT_BYTE_NO_ACK;
      if (event_master != NULL) event_master(state_master);
      break;
      
      
    /* arbitration lost */
    case TW_MT_ARB_LOST:
      //    case TW_MR_ARB_LOST:
      state_master = I2CM_LOST_ARBITRATION;
      if (event_master != NULL) event_master(state_master);
      TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE) ;
      break;

    /* arbitration lost and addressed */
    case TW_SR_ARB_LOST_SLA_ACK:
    case TW_SR_ARB_LOST_GCALL_ACK:
      state_master = I2CM_LOST_ARBITRATION;
      state_slave = I2CS_ADRESSED_RECEIVE;
      if (event_master != NULL) event_master(state_master);
      if (event_slave != NULL) event_slave(state_slave);
      /* FIXME : update TWCR ? */
      break;
      
    case TW_ST_ARB_LOST_SLA_ACK:
      state_master = I2CM_LOST_ARBITRATION;
      state_slave = I2CS_ADRESSED_SEND;
      if (event_master != NULL) event_master(state_master);
      if (event_slave != NULL) event_slave(state_slave);
      /* FIXME : update TWCR ? */
      break;
      

    /* slave */
    
    /* the slave has been adressed */
    case TW_ST_SLA_ACK: 
      state_slave = I2CS_ADRESSED_SEND;
      if (event_slave != NULL) event_slave(state_slave);
      TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA) | (1<<TWIE) ;
      break;
      
    case TW_SR_GCALL_ACK:
    case TW_SR_SLA_ACK:
      state_slave = I2CS_ADRESSED_RECEIVE;
      if (event_slave != NULL) event_slave(state_slave);
      TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA) | (1<<TWIE) ;
      break;

    case TW_ST_LAST_DATA:
      state_slave = I2CS_READY;
      TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA) | (1<<TWIE) ;
      break;

    /* transmitting data */
    case TW_ST_DATA_ACK:
      state_slave = I2CS_SENT_BYTE;
      if (event_slave != NULL) event_slave(state_slave);
      break;

    case TW_ST_DATA_NACK:
      state_slave = I2CS_SENT_BYTE_NO_ACK;
      if (event_slave != NULL) event_slave(state_slave);
      break;
            
    case TW_SR_GCALL_DATA_NACK:
    case TW_SR_DATA_NACK:    
    case TW_SR_GCALL_DATA_ACK:
    case TW_SR_DATA_ACK:
      received_data = TWDR;
      state_slave = I2CS_RECEIVED_BYTE;
      if (event_slave != NULL) event_slave(state_slave);
      break;
    
    case TW_SR_STOP:
      state_slave = I2CS_READY;
      if (event_slave != NULL) event_slave(state_slave);
      TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA) | (1<<TWIE) ;
      break;
    
    
    /* error >> send stop */
    case TW_BUS_ERROR:
      sbi(TWCR, TWSTO);
      sbi(TWCR, TWINT);
      break;
    
     /* default */
    default :
      sbi(TWCR, TWINT);
      break;

    }
}








